use craftflow_nbt::DynNBT;
use enum_index::EnumIndex;
use enum_index_derive::EnumIndex;
use rust_mc_proto::{DataReader, DataWriter, Packet};
use uuid::Uuid;

use crate::ServerError;

use super::{
	DyeColor, IdOr, IdSet, Property, ReadWriteAny, component::TextComponent, sound::SoundEvent,
};

pub const SLOT_COMPONENT_LENGTH: u16 = 96;

#[derive(Clone)]
pub struct JukeboxSong {
	pub sound: IdOr<SoundEvent>,
	pub description: TextComponent,
	/// The duration the songs should play for, in seconds.
	pub duration: f32,
	/// The output strength given by a comparator. Between 0 and 15.
	pub output: i32,
}

#[derive(Clone)]
pub enum JukeboxPlayable {
	ByName(String),
	ById(IdOr<JukeboxSong>),
}

#[derive(Clone)]
pub enum ChickenVariant {
	ByName(String),
	ById(u32),
}

#[derive(Clone)]
pub struct PaintingVariant {
	pub width: i32,
	pub height: i32,
	/// The texture for the painting. The Notchian client uses the corresponding asset located at textures/painting.
	pub asset_id: String,
	pub title: Option<TextComponent>,
	pub author: Option<TextComponent>,
}

#[derive(Clone)]
pub enum ProvidesTrimMaterial {
	ByName(String),
	ById(IdOr<TrimMaterial>),
}

#[derive(Clone)]
pub enum BlockPredicatePropertyMatch {
	Exact(String),
	Range(String, String),
}

#[derive(Clone)]
pub struct BlockPredicateExactDataMatcher {
	/// ID of the data component as listed in the table of data component types above.
	pub component_index: usize,
	pub value: StructuredComponent,
}

#[derive(Clone)]
pub struct BlockPredicatePartialDataMatcher {
	/// 0: damage, 1: enchantments, 2: stored_enchantments, 3: potion_contents, 4: custom_data, 5: container, 6: bundle_contents, 7: firework_explosion, 8: fireworks, 9: writable_book_content, 10: written_book_content, 11: attribute_modifiers, 12: trim, 13: jukebox_playable.
	pub type_id: u8,
	/// https://minecraft.wiki/w/Data_component_predicate
	pub predicate: DynNBT,
}

#[derive(Clone)]
pub struct BlockPredicateProperty {
	pub name: String,
	pub matches: BlockPredicatePropertyMatch,
}

#[derive(Clone)]
pub struct BlockPredicate {
	pub blocks: Option<IdSet>,
	pub properties: Option<Vec<BlockPredicateProperty>>,
	pub nbt: Option<DynNBT>,
	pub exact_matchers: Vec<BlockPredicateExactDataMatcher>,
	pub partial_matchers: Vec<BlockPredicatePartialDataMatcher>,
}

#[derive(Clone)]
pub struct Instrument {
	pub sound: IdOr<SoundEvent>,
	pub sound_range: f32,
	pub range: f32,
	pub description: TextComponent,
}

#[derive(Clone)]
pub enum ConsumeEffect {
	/// Effects, Probability
	ApplyEffects(Vec<PotionEffect>, f32),
	RemoveEffects(IdSet),
	ClearAllEffects,
	/// Diameter
	TeleportRandomly(f32),
	PlaySound(SoundEvent),
}

impl ReadWriteAny<ConsumeEffect> for Packet {
	fn read_any(&mut self) -> Result<ConsumeEffect, ServerError> {
		match self.read_varint()? {
			0 => {
				let len = self.read_usize_varint()?;
				let mut effects = Vec::new();
				for _ in 0..len {
					effects.push(self.read_any()?);
				}
				let probability = self.read_float()?;
				Ok(ConsumeEffect::ApplyEffects(effects, probability))
			}
			1 => Ok(ConsumeEffect::RemoveEffects(self.read_any()?)),
			2 => Ok(ConsumeEffect::ClearAllEffects),
			3 => Ok(ConsumeEffect::TeleportRandomly(self.read_float()?)),
			4 => Ok(ConsumeEffect::PlaySound(self.read_any()?)),
			_ => Err(ServerError::WrongPacket),
		}
	}
	fn write_any(&mut self, val: &ConsumeEffect) -> Result<(), ServerError> {
		match val {
			ConsumeEffect::ApplyEffects(effects, probability) => {
				self.write_usize_varint(effects.len())?;
				for effect in effects {
					self.write_any(effect)?;
				}
				self.write_float(*probability)?;
			}
			ConsumeEffect::RemoveEffects(set) => {
				self.write_any(set)?;
			}
			ConsumeEffect::ClearAllEffects => {}
			ConsumeEffect::TeleportRandomly(diameter) => {
				self.write_float(*diameter)?;
			}
			ConsumeEffect::PlaySound(sound) => {
				self.write_any(sound)?;
			}
		}
		Ok(())
	}
}

#[derive(Clone)]
pub struct TrimMaterial {
	pub suffix: String,
	/// Key - Armor Material Type Identifier
	/// Value - Overriden Asset Name
	pub overrides: Vec<(String, String)>,
	pub description: TextComponent,
}

impl ReadWriteAny<TrimMaterial> for Packet {
	fn read_any(&mut self) -> Result<TrimMaterial, ServerError> {
		Ok(TrimMaterial {
			suffix: self.read_string()?,
			overrides: {
				let len = self.read_usize_varint()?;
				let mut overrides = Vec::new();
				for _ in 0..len {
					overrides.push((self.read_string()?, self.read_string()?));
				}
				overrides
			},
			description: self.read_any()?,
		})
	}
	fn write_any(&mut self, val: &TrimMaterial) -> Result<(), ServerError> {
		self.write_string(&val.suffix)?;
		self.write_usize_varint(val.overrides.len())?;
		for ele in &val.overrides {
			self.write_string(&ele.0)?;
			self.write_string(&ele.1)?;
		}
		self.write_any(&val.description)?;
		Ok(())
	}
}

#[derive(Clone)]
pub struct TrimPattern {
	pub asset_name: String,
	pub template_item: i32,
	pub description: TextComponent,
	pub decal: bool,
}

impl ReadWriteAny<TrimPattern> for Packet {
	fn read_any(&mut self) -> Result<TrimPattern, ServerError> {
		Ok(TrimPattern {
			asset_name: self.read_string()?,
			template_item: self.read_varint()?,
			description: self.read_any()?,
			decal: self.read_boolean()?,
		})
	}
	fn write_any(&mut self, val: &TrimPattern) -> Result<(), ServerError> {
		self.write_string(&val.asset_name)?;
		self.write_varint(val.template_item)?;
		self.write_any(&val.description)?;
		self.write_boolean(val.decal)?;
		Ok(())
	}
}

#[derive(Clone)]
pub struct PotionEffectDetail {
	pub amplifier: i32,
	/// -1 for infinite.
	pub duration: i32,
	pub ambient: bool,
	pub show_particles: bool,
	pub show_icon: bool,
	pub hidden_effect: Option<Box<PotionEffectDetail>>,
}

impl<T: DataReader + DataWriter> ReadWriteAny<PotionEffectDetail> for T {
	fn read_any(&mut self) -> Result<PotionEffectDetail, ServerError> {
		Ok(PotionEffectDetail {
			amplifier: self.read_varint()?,
			duration: self.read_varint()?,
			ambient: self.read_boolean()?,
			show_particles: self.read_boolean()?,
			show_icon: self.read_boolean()?,
			hidden_effect: if self.read_boolean()? {
				Some(Box::new(self.read_any()?))
			} else {
				None
			},
		})
	}
	fn write_any(&mut self, val: &PotionEffectDetail) -> Result<(), ServerError> {
		self.write_varint(val.amplifier)?;
		self.write_varint(val.duration)?;
		self.write_boolean(val.ambient)?;
		self.write_boolean(val.show_particles)?;
		self.write_boolean(val.show_icon)?;
		self.write_boolean(val.hidden_effect.is_some())?;
		if let Some(val) = &val.hidden_effect {
			self.write_any(&val.as_ref().clone())?;
		}
		Ok(())
	}
}

#[derive(Clone)]
pub struct PotionEffect {
	pub type_id: u32,
	pub detail: PotionEffectDetail,
}

impl<T: DataReader + DataWriter> ReadWriteAny<PotionEffect> for T {
	fn read_any(&mut self) -> Result<PotionEffect, ServerError> {
		Ok(PotionEffect {
			type_id: self.read_u32_varint()?,
			detail: self.read_any()?,
		})
	}
	fn write_any(&mut self, val: &PotionEffect) -> Result<(), ServerError> {
		self.write_u32_varint(val.type_id)?;
		self.write_any(&val.detail)?;
		Ok(())
	}
}

#[derive(Clone)]
pub struct FireworkExplosion {
	/// Can be one of the following:
	/// - 0 - Small ball
	/// - 1 - Large ball
	/// - 2 - Star
	/// - 3 - Creeper
	/// - 4 - Burst
	pub shape: u8,
	pub colors: Vec<i32>,
	pub fade_colors: Vec<i32>,
	pub has_trail: bool,
	pub has_twinkle: bool,
}

impl<T: DataReader + DataWriter> ReadWriteAny<FireworkExplosion> for T {
	fn read_any(&mut self) -> Result<FireworkExplosion, ServerError> {
		Ok(FireworkExplosion {
			shape: self.read_u8_varint()?,
			colors: {
				let len = self.read_usize_varint()?;
				let mut colors = Vec::new();
				for _ in 0..len {
					colors.push(self.read_int()?);
				}
				colors
			},
			fade_colors: {
				let len = self.read_usize_varint()?;
				let mut colors = Vec::new();
				for _ in 0..len {
					colors.push(self.read_int()?);
				}
				colors
			},
			has_trail: self.read_boolean()?,
			has_twinkle: self.read_boolean()?,
		})
	}
	fn write_any(&mut self, val: &FireworkExplosion) -> Result<(), ServerError> {
		self.write_u8_varint(val.shape)?;
		self.write_usize_varint(val.colors.len())?;
		for color in val.colors.clone() {
			self.write_int(color)?;
		}
		self.write_usize_varint(val.fade_colors.len())?;
		for color in val.fade_colors.clone() {
			self.write_int(color)?;
		}
		self.write_boolean(val.has_trail)?;
		self.write_boolean(val.has_twinkle)?;
		Ok(())
	}
}

#[derive(Clone)]
pub struct HiveBee {
	pub entity_data: DynNBT,
	pub ticks_in_hive: i32,
	pub min_ticks_in_hive: i32,
}

impl ReadWriteAny<HiveBee> for Packet {
	fn read_any(&mut self) -> Result<HiveBee, ServerError> {
		Ok(HiveBee {
			entity_data: self.read_any()?,
			ticks_in_hive: self.read_varint()?,
			min_ticks_in_hive: self.read_varint()?,
		})
	}
	fn write_any(&mut self, val: &HiveBee) -> Result<(), ServerError> {
		self.write_any(&val.entity_data)?;
		self.write_varint(val.ticks_in_hive)?;
		self.write_varint(val.min_ticks_in_hive)?;
		Ok(())
	}
}

#[derive(Clone)]
pub enum BannerLayer {
	Direct {
		asset_id: String,
		translation_key: String,
		color: DyeColor,
	},
	Registry {
		pattern_id: i32,
		color: DyeColor,
	},
}

impl<T: DataReader + DataWriter> ReadWriteAny<BannerLayer> for T {
	fn read_any(&mut self) -> Result<BannerLayer, ServerError> {
		match self.read_varint()? {
			0 => Ok(BannerLayer::Direct {
				asset_id: self.read_string()?,
				translation_key: self.read_string()?,
				color: self.read_any()?,
			}),
			id => Ok(BannerLayer::Registry {
				pattern_id: id - 1,
				color: self.read_any()?,
			}),
		}
	}
	fn write_any(&mut self, val: &BannerLayer) -> Result<(), ServerError> {
		match val {
			BannerLayer::Direct {
				asset_id,
				translation_key,
				color,
			} => {
				self.write_varint(0)?;
				self.write_string(&asset_id)?;
				self.write_string(&translation_key)?;
				self.write_any(color)?;
			}
			BannerLayer::Registry { pattern_id, color } => {
				self.write_varint(pattern_id + 1)?;
				self.write_any(color)?;
			}
		}
		Ok(())
	}
}

#[derive(Clone)]
pub struct AttributeModifier {
	pub attribute_id: u32,
	pub modifier_id: String,
	pub value: f64,
	/// The operation to be applied upon the value. Can be one of the following:
	/// - 0 - Add
	/// - 1 - Multiply base
	/// - 2 - Multiply total
	pub operation: u8,
	/// The item slot placement required for the modifier to have effect.
	/// Can be one of the following:
	/// - 0 - Any
	/// - 1 - Main hand
	/// - 2 - Off hand
	/// - 3 - Hand
	/// - 4 - Feet
	/// - 5 - Legs
	/// - 6 - Chest
	/// - 7 - Head
	/// - 8 - Armor
	/// - 9 - Body
	pub slot: u8,
}

impl<T: DataReader + DataWriter> ReadWriteAny<AttributeModifier> for T {
	fn read_any(&mut self) -> Result<AttributeModifier, ServerError> {
		Ok(AttributeModifier {
			attribute_id: self.read_u32_varint()?,
			modifier_id: self.read_string()?,
			value: self.read_double()?,
			operation: self.read_u8_varint()?,
			slot: self.read_u8_varint()?,
		})
	}
	fn write_any(&mut self, val: &AttributeModifier) -> Result<(), ServerError> {
		self.write_u32_varint(val.attribute_id)?;
		self.write_string(&val.modifier_id)?;
		self.write_double(val.value)?;
		self.write_u8_varint(val.operation)?;
		self.write_u8_varint(val.slot)?;
		Ok(())
	}
}

#[derive(Clone)]
pub struct ToolRule {
	pub blocks: IdSet,
	pub speed: Option<f32>,
	pub correct_drop_for_blocks: Option<bool>,
}

impl<T: DataReader + DataWriter> ReadWriteAny<ToolRule> for T {
	fn read_any(&mut self) -> Result<ToolRule, ServerError> {
		Ok(ToolRule {
			blocks: self.read_any()?,
			speed: if self.read_boolean()? {
				Some(self.read_float()?)
			} else {
				None
			},
			correct_drop_for_blocks: if self.read_boolean()? {
				Some(self.read_boolean()?)
			} else {
				None
			},
		})
	}
	fn write_any(&mut self, val: &ToolRule) -> Result<(), ServerError> {
		self.write_any(&val.blocks)?;
		if let Some(speed) = val.speed {
			self.write_boolean(true)?;
			self.write_float(speed)?;
		} else {
			self.write_boolean(false)?;
		}
		if let Some(cdfb) = val.correct_drop_for_blocks {
			self.write_boolean(true)?;
			self.write_boolean(cdfb)?;
		} else {
			self.write_boolean(false)?;
		}
		Ok(())
	}
}

#[derive(Clone)]
pub struct DamageReduction {
	pub horizontal_blocking_angle: f32,
	pub damage_kind: Option<IdSet>,
	pub base: f32,
	pub factor: f32,
}

impl<T: DataReader + DataWriter> ReadWriteAny<DamageReduction> for T {
	fn read_any(&mut self) -> Result<DamageReduction, ServerError> {
		Ok(DamageReduction {
			horizontal_blocking_angle: self.read_float()?,
			damage_kind: if self.read_boolean()? {
				Some(self.read_any()?)
			} else {
				None
			},
			base: self.read_float()?,
			factor: self.read_float()?,
		})
	}
	fn write_any(&mut self, val: &DamageReduction) -> Result<(), ServerError> {
		self.write_float(val.horizontal_blocking_angle)?;
		if let Some(id_set) = val.damage_kind.clone() {
			self.write_boolean(true)?;
			self.write_any(&id_set)?;
		} else {
			self.write_boolean(false)?;
		}
		self.write_float(val.base)?;
		self.write_float(val.factor)?;
		Ok(())
	}
}

/// https://minecraft.wiki/w/Java_Edition_protocol/Slot_data#Structured_components
#[derive(Clone, EnumIndex)]
pub enum StructuredComponent {
	CustomData(DynNBT), // 0
	/// 1 - 99
	MaxStackSize(i32), // 1
	MaxDamage(i32),     // 2
	Damage(i32),        // 3
	Unbreakable,        // 4
	CustomName(TextComponent), // 5
	ItemName(TextComponent), // 6
	ItemModel(String),  // 7
	Lore(Vec<TextComponent>), // 8
	/// Can be one of the following:
	/// - 0 - Common (white)
	/// - 1 - Uncommon (yellow)
	/// - 2 - Rare (aqua)
	/// - 3 - Epic (pink)
	Rarity(u8), // 9
	/// Key: The ID of the enchantment in the enchantment registry. \
	/// Value: The level of the enchantment.
	Enchantments(Vec<(u32, i32)>), // 10
	CanPlaceOn(Vec<BlockPredicate>), // 11
	CanBreak(Vec<BlockPredicate>), // 12
	AttributeModifiers(Vec<AttributeModifier>), // 13
	/// Floats, Flags, Strings, Colors
	CustomModelData(Vec<f32>, Vec<bool>, Vec<String>, Vec<i32>), // 14
	/// Parameters:
	/// - Hide Tooltip
	/// - The IDs of data components in the minecraft:data_component_type registry to hide.
	TooltipDisplay(bool, Vec<u64>), // 15
	/// Accumulated anvil usage cost.
	/// The client displays "Too Expensive" if the value is greater than 40 and
	/// the player is not in creative mode (more specifically,
	/// if they don't have the insta-build flag enabled).
	/// This behavior can be overridden by setting the level
	/// with the Set Container Property packet.
	RepairCost(i32), // 16
	/// Marks the item as non-interactive on the creative inventory (the first 5 rows of items). \
	/// This is used internally by the client on the paper icon in the saved hot-bars tab.
	CreativeSlotLock, // 17
	EnchantmentGlintOverride(bool), // 18
	/// Marks the projectile as intangible (cannot be picked-up).
	IntangibleProjectile, // 19
	/// Parameters:
	/// - Nutrition, Non-negative
	/// - How much saturation will be given after consuming the item.
	/// - Whether the item can always be eaten, even at full hunger.
	Food(u32, f32, bool), // 20
	/// Animation: 0: none, 1: eat, 2: drink, 3: block, 4: bow, 5: spear, 6: crossbow, 7: spyglass, 8: toot_horn, 9: brush
	Consumable {
		consume_seconds: f32,
		animation: u8,
		sound: IdOr<SoundEvent>,
		has_particles: bool,
		effects: Vec<ConsumeEffect>,
	}, // 21
	UseRemainder(Slot), // 22
	/// Group of items to apply the cooldown to. Only present if Has cooldown group is true; otherwise defaults to the item's identifier.
	UseCooldown {
		seconds: f32,
		group: Option<String>,
	}, // 23
	/// Parameter - Types, Tag specifying damage types the item is immune to. Not prefixed by '#'!.
	DamageResistant(String), // 24
	Tool {
		rules: Vec<ToolRule>,
		default_mining_speed: f32,
		damage_per_block: i32,
	}, // 25
	Weapon {
		damage_per_attack: i32,
		disable_blocking_for_seconds: f32,
	}, // 26
	// Opaque internal value controlling how expensive enchantments may be offered.
	Enchantable(i32), // 27
	Equippable {
		slot: u8,
		equip_sound: IdOr<SoundEvent>,
		model: Option<String>,
		camera_overlay: Option<String>,
		allowed_entities: Option<IdSet>,
		dispensable: bool,
		swappable: bool,
		damage_on_hurt: bool,
	}, // 28
	/// Items that can be combined with this item in an anvil to repair it.
	Repairable(IdSet), // 29
	/// Makes the item function like elytra.
	Glider, // 30
	TooltipStyle(String), // 31
	/// Makes the item function like a totem of undying.
	DeathProtection(Vec<ConsumeEffect>), // 32
	BlockAttacks {
		block_delay_seconds: f32,
		disable_cooldown_scale: f32,
		damage_reductions: Vec<DamageReduction>,
		item_damage_threshold: f32,
		item_damage_base: f32,
		item_damage_factor: f32,
		bypassed_by: Option<String>,
		block_sound: Option<IdOr<SoundEvent>>,
		disable_sound: Option<IdOr<SoundEvent>>,
	}, // 33
	/// The enchantments stored in this enchanted book.
	/// Key: The ID of the enchantment in the enchantment registry. \
	/// Value: The level of the enchantment.
	StoredEnchantments(Vec<(u64, i32)>), // 34
	DyedColor(i32),   // 35
	MapColor(i32),    // 36
	MapId(i32),       // 37
	MapDecorations(DynNBT), // 38
	/// Type of post processing. Can be either:
	/// - 0 - Lock
	/// - 1 - Scale
	MapPostProcessing(u8), // 39
	/// Projectiles loaded into a charged crossbow.
	ChargedProjectiles(Vec<Slot>), // 40
	BundleContents(Vec<Slot>), // 41
	PotionContents {
		potion_id: Option<u64>,
		custom_color: Option<i32>,
		custom_effects: Vec<PotionEffect>,
		custom_name: String,
	}, // 42
	/// Parameter - Effect Multiplier
	PotionDurationScale(f32), // 43
	/// Key - The ID of the effect in the potion effect type registry.
	/// Value - The duration of the effect.
	SuspiciousStewEffects(Vec<(u64, i32)>), // 44
	/// Parameter - Pages
	/// Page:
	/// - The raw text of the page
	/// - The content after passing through chat filters
	WritableBookContent(Vec<(String, Option<String>)>), // 45
	WrittenBookContent {
		raw_title: String,
		filtered_title: Option<String>,
		author: String,
		generation: i32,
		/// Page:
		/// - The raw text of the page
		/// - The content after passing through chat filters
		pages: Vec<(String, Option<String>)>,
		resolved: bool,
	}, // 46
	/// Armor's trim pattern and color
	Trim(IdOr<TrimMaterial>, IdOr<TrimPattern>), // 47
	DebugStrickState(DynNBT), // 48
	EntityData(DynNBT), // 49
	BucketEntityData(DynNBT), // 50
	BlockEntityData(DynNBT), // 51
	Instrument(IdOr<Instrument>), // 52
	ProvidesTrimMaterial(ProvidesTrimMaterial), // 53
	/// Between 0 and 4.
	OminousBottleAmplifier(u8), // 54
	JukeboxPlayable(JukeboxPlayable), // 55
	/// A pattern identifier like #minecraft:pattern_item/globe
	ProvidesBannerPatterns(String), // 56
	Recipes(DynNBT),  // 57
	LodestoneTracker {
		has_global_position: bool,
		dimension: String,
		position: (f64, f64, f64),
		tracked: bool,
	}, // 58
	FireworkExplosion(FireworkExplosion), // 59
	Fireworks {
		flight_duration: i32,
		explosions: Vec<FireworkExplosion>,
	}, // 60
	Profile {
		name: Option<String>,
		unique_id: Option<Uuid>,
		properties: Vec<Property>,
	}, // 61
	NoteBlockSound(String), // 62
	BannerPatterns(Vec<BannerLayer>), // 63
	BaseColor(DyeColor), // 64
	/// The ID of the items in the item registry.
	PotDecorations([u64; 4]), // 65
	/// Items inside a container of any type.
	Container(Vec<Slot>), // 66
	BlockState(Vec<(String, String)>), // 67
	Bees(Vec<HiveBee>), // 68
	Lock(String),     // 69
	ContainerLoot(DynNBT), // 70
	BreakSound(IdOr<SoundEvent>), // 71
	VillagerVariant(u64), // 72
	WolfVariant(u64), // 73
	WolfSoundVariant(u64), // 74
	WolfCollar(u8),   // 75
	/// 0: red, 1: snow
	FoxVariant(u8), // 76
	/// 0: small, 1: medium, 2: large.
	SalmonSize(u8), // 77
	ParrotVariant(u64), // 78
	/// 0: kob, 1: sunstreak, 2: snooper, 3: dasher, 4: brinely, 5: spotty, 6: flopper, 7: stripey, 8: glitter, 9: blockfish, 10: betty, 11: clayfish.
	TropicalFishPattern(u8), // 79
	TropicalFishBaseColor(DyeColor), // 80
	TropicalFishPatternColor(DyeColor), // 81
	/// 0: red, 1: brown.
	MooshroomVariant(u8), // 82
	/// 0: brown, 1: white, 2: black, 3: white splotched, 4: gold, 5: salt, 6: evil.
	RabbitVariant(u8), // 83
	PigVariant(u64),  // 84
	CowVariant(u64),  // 85
	ChickenVariant(ChickenVariant), // 86
	FrogVariant(u64), // 87
	/// 0: white, 1: creamy, 2: chestnut, 3: brown, 4: black, 5: gray, 6: dark brown.
	HorseVariant(u8), // 88
	PaintingVariant(PaintingVariant), // 89
	/// 0: creamy, 1: white, 2: brown, 3: gray.
	LlamaVariant(u8), // 90
	/// 0: lucy, 1: wild, 2: gold, 3: cyan, 4: blue.
	AxolotlVariant(u8), // 91
	CatVariant(u64),  // 92
	CatCollar(DyeColor), // 93
	SheepColor(DyeColor), // 94
	ShulkerColor(DyeColor), // 95
}

impl ReadWriteAny<StructuredComponent> for Packet {
	fn read_any(&mut self) -> Result<StructuredComponent, ServerError> {
		let id = self.read_u16_varint()?; // id

		match id {
			0 => Ok(StructuredComponent::CustomData(self.read_any()?)),
			1 => Ok(StructuredComponent::MaxStackSize(self.read_varint()?)),
			2 => Ok(StructuredComponent::MaxDamage(self.read_varint()?)),
			3 => Ok(StructuredComponent::Unbreakable),
			4 => Ok(StructuredComponent::CustomName(self.read_any()?)),
			5 => Ok(StructuredComponent::ItemName(self.read_any()?)),
			6 => Ok(StructuredComponent::ItemModel(self.read_string()?)),
			7 => {
				let len: usize = self.read_usize_varint()?;
				let mut lore = Vec::new();

				for _ in 0..len {
					lore.push(self.read_any()?);
				}

				Ok(StructuredComponent::Lore(lore))
			}
			8 => Ok(StructuredComponent::Rarity(self.read_u8_varint()?)),
			9 => {
				let len: usize = self.read_usize_varint()?;
				let mut enchantments = Vec::new();

				for _ in 0..len {
					enchantments.push((self.read_u32_varint()?, self.read_varint()?));
				}

				Ok(StructuredComponent::Enchantments(enchantments))
			}
			_ => Err(ServerError::UnexpectedSlotComponent),
		}
	}

	fn write_any(&mut self, val: &StructuredComponent) -> Result<(), ServerError> {
		self.write_usize_varint(val.enum_index())?;

		match val {
			StructuredComponent::CustomData(val) => self.write_any(val)?,
			StructuredComponent::MaxStackSize(val) => self.write_varint(*val)?,
			StructuredComponent::MaxDamage(val) => self.write_varint(*val)?,
			StructuredComponent::Unbreakable => {}
			StructuredComponent::CustomName(val) => self.write_any(val)?,
			StructuredComponent::ItemName(val) => self.write_any(val)?,
			StructuredComponent::ItemModel(val) => self.write_string(val)?,
			StructuredComponent::Lore(val) => {
				self.write_usize_varint(val.len())?;

				for text in val {
					self.write_any(text)?;
				}
			}
			StructuredComponent::Rarity(val) => self.write_u8_varint(*val)?,
			StructuredComponent::Enchantments(val) => {
				self.write_usize_varint(val.len())?;

				for (id, lvl) in val {
					self.write_u32_varint(*id)?;
					self.write_varint(*lvl)?;
				}
			}
			_ => {}
		}

		Ok(())
	}
}

#[derive(Clone)]
pub struct Slot {
	pub id: i32,
	pub amount: i32,
	pub components: Vec<StructuredComponent>,
}

impl ReadWriteAny<Option<Slot>> for Packet {
	fn read_any(&mut self) -> Result<Option<Slot>, ServerError> {
		let amount = self.read_varint()?;

		if amount > 0 {
			let id = self.read_varint()?;
			let components_len = self.read_varint()?;
			self.read_varint()?; // components_to_remove_len
			let mut components = Vec::new();
			for _ in 0..components_len {
				components.push(self.read_any()?);
			}

			Ok(Some(Slot {
				id,
				amount,
				components,
			}))
		} else {
			Ok(None)
		}
	}

	fn write_any(&mut self, val: &Option<Slot>) -> Result<(), ServerError> {
		if let Some(val) = val {
			self.write_varint(val.amount)?;
			self.write_varint(val.id)?;
			self.write_usize_varint(val.components.len())?;
			let mut components_to_remove: Vec<u16> = (0..SLOT_COMPONENT_LENGTH).collect();
			self.write_usize_varint(SLOT_COMPONENT_LENGTH as usize - val.components.len())?;
			for comp in val.components.clone() {
				let id = comp.enum_index() as u16;
				self.write_any(&comp)?;
				if let Some(index) = components_to_remove.iter().position(|v| *v == id) {
					components_to_remove.swap_remove(index);
				}
			}
			for id in components_to_remove {
				self.write_u16_varint(id)?;
			}
		} else {
			self.write_varint(0)?;
		}
		Ok(())
	}
}

#[derive(Clone)]
pub struct HashedSlot {
	pub id: i32,
	pub amount: i32,
	/// id -> crc32 hash
	pub components: Vec<(u16, i32)>,
}

impl<T: DataReader + DataWriter> ReadWriteAny<Option<HashedSlot>> for T {
	fn read_any(&mut self) -> Result<Option<HashedSlot>, ServerError> {
		let amount = self.read_varint()?;

		if amount > 0 {
			let id = self.read_varint()?;
			let components_len = self.read_varint()?;
			self.read_varint()?; // components_to_remove_len
			let mut components = Vec::new();
			for _ in 0..components_len {
				components.push((self.read_u16_varint()?, self.read_int()?));
			}

			Ok(Some(HashedSlot {
				id,
				amount,
				components,
			}))
		} else {
			Ok(None)
		}
	}

	fn write_any(&mut self, val: &Option<HashedSlot>) -> Result<(), ServerError> {
		if let Some(val) = val {
			self.write_varint(val.amount)?;
			self.write_varint(val.id)?;
			self.write_usize_varint(val.components.len())?;
			let mut components_to_remove: Vec<u16> = (0..SLOT_COMPONENT_LENGTH).collect();
			self.write_usize_varint(SLOT_COMPONENT_LENGTH as usize - val.components.len())?;
			for (id, hash) in val.components.clone() {
				self.write_u16_varint(id)?;
				self.write_int(hash)?;
				if let Some(index) = components_to_remove.iter().position(|v| *v == id) {
					components_to_remove.swap_remove(index);
				}
			}
			for id in components_to_remove {
				self.write_u16_varint(id)?;
			}
		} else {
			self.write_varint(0)?;
		}
		Ok(())
	}
}
